@@ -1,7 +1,7 @@
 (**************************************************************************)
 (*                                BELENIOS                                *)
 (*                                                                        *)
-(*  Copyright © 2012-2016 Inria                                           *)
+(*  Copyright © 2012-2018 Inria                                           *)
 (*                                                                        *)
 (*  This program is free software: you can redistribute it and/or modify  *)
 (*  it under the terms of the GNU Affero General Public License as        *)
@@ -20,10 +20,29 @@
 (**************************************************************************)
 
 open Platform
-open Serializable_t
+open Serializable_j
 open Signatures
 open Common
 
+(** Parsing helpers *)
+
+let of_string x =
+  let params = params_of_string Yojson.Safe.read_json x in
+  {
+    e_params = params;
+    e_fingerprint = sha256_b64 x;
+  }
+
+let get_group x =
+  let w = Yojson.Safe.to_string x.e_params.e_public_key in
+  let module W = (val Group.wrapped_pubkey_of_string w) in
+  let module X =
+    struct
+      module G = W.G
+      let election = {x with e_params = {x.e_params with e_public_key = W.y}}
+    end
+  in (module X : ELECTION_DATA)
+
 (** Helper functions *)
 
 let check_modulo p x = Z.(geq x zero && lt x p)
@@ -33,83 +52,19 @@ let question_length q =
                              | Some true -> 1
                              | _ -> 0
 
-(** Simple monad *)
-
-module MakeSimpleMonad (G : GROUP) = struct
-  type 'a t = unit -> 'a
-  let ballots = ref []
-  let return x () = x
-  let bind x f = f (x ())
-  let fail e = raise e
-
-  let prng = lazy (pseudo_rng (random_string secure_rng 16))
-
-  let random q =
-    let size = Z.bit_length q / 8 + 1 in
-    fun () ->
-      let r = random_string (Lazy.force prng) size in
-      Z.(of_bits r mod q)
-
-  type elt = G.t ballot
-  let cast x () = ballots := x :: !ballots
-  let fold f x () = List.fold_left (fun accu b -> f () b accu ()) x !ballots
-  let cardinal () = List.length !ballots
-end
-
-(** Distributed key generation *)
-
-module MakeSimpleDistKeyGen (G : GROUP) (M : RANDOM) = struct
-  open G
-  open M
-
-  let ( >>= ) = bind
-  let ( / ) x y = x *~ invert y
-
-  (** Fiat-Shamir non-interactive zero-knowledge proofs of
-      knowledge *)
-
-  let fs_prove gs x oracle =
-    random q >>= fun w ->
-    let commitments = Array.map (fun g -> g **~ w) gs in
-    let challenge = oracle commitments in
-    let response = Z.((w + x * challenge) mod q) in
-    return {challenge; response}
-
-  let generate_and_prove () =
-    random q >>= fun x ->
-    let trustee_public_key = g **~ x in
-    let zkp = "pok|" ^ G.to_string trustee_public_key ^ "|" in
-    fs_prove [| g |] x (G.hash zkp) >>= fun trustee_pok ->
-    return (x, {trustee_pok; trustee_public_key})
-
-  let check {trustee_pok; trustee_public_key = y} =
-    G.check y &&
-    let {challenge; response} = trustee_pok in
-    check_modulo q challenge &&
-    check_modulo q response &&
-    let commitment = g **~ response / (y **~ challenge) in
-    let zkp = "pok|" ^ G.to_string y ^ "|" in
-    Z.(challenge =% G.hash zkp [| commitment |])
-
-  let combine pks =
-    Array.fold_left (fun y {trustee_public_key; _} ->
-      y *~ trustee_public_key
-    ) G.one pks
-
-end
-
 (** Homomorphic elections *)
 
-module MakeElection (G : GROUP) (M : RANDOM) = struct
-  open G
-
+module Make (W : ELECTION_DATA) (M : RANDOM) = struct
   type 'a m = 'a M.t
   open M
   let ( >>= ) = bind
 
-  type elt = G.t
+  type elt = W.G.t
+
+  module G = W.G
+  open G
+  let election = W.election
 
-  type t = elt election
   type private_key = Z.t
   type public_key = elt
 
@@ -130,9 +85,9 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
       beta = c1.beta *~ c2.beta;
     }
 
-  let neutral_ciphertext e = Array.map (fun q ->
+  let neutral_ciphertext () = Array.map (fun q ->
     Array.make (question_length q) dummy_ciphertext
-  ) e.e_params.e_questions
+  ) election.e_params.e_questions
 
   let combine_ciphertexts = Array.mmap2 eg_combine
 
@@ -142,9 +97,11 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
 
   (** ElGamal encryption. *)
   let eg_encrypt y r x =
+    (* FIXME: side channel *)
+    let g' = if x = 0 then G.one else g **~ Z.of_int x in
     {
       alpha = g **~ r;
-      beta = y **~ r *~ g **~ Z.of_int x;
+      beta = y **~ r *~ g';
     }
 
   let dummy_proof =
@@ -268,7 +225,8 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
             if i <> index_true then (
               random q >>= fun challenge ->
               random q >>= fun response ->
-              let nbeta = cS.beta / (g **~ Z.of_int (min+i-1)) in
+              let g' = if min+i-1 = 0 then G.one else g **~ Z.of_int (min+i-1) in
+              let nbeta = cS.beta / g' in
               let j = 2*i in
               overall_proof.(i) <- {challenge; response};
               commitments.(j) <- g **~ response *~ cS.alpha **~ challenge;
@@ -323,7 +281,8 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
           if i < max-min+2 then (
             random q >>= fun challenge ->
             random q >>= fun response ->
-            let nbeta = cS.beta / (g **~ Z.of_int (min+i-1)) in
+            let g' = if min+i-1 = 0 then G.one else g **~ Z.of_int (min+i-1) in
+            let nbeta = cS.beta / g' in
             let j = 2*i in
             overall_proof.(i) <- {challenge; response};
             commitments.(j) <- g **~ response *~ cS.alpha **~ challenge;
@@ -397,7 +356,8 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
             let {challenge; response} = overall_proof.(i) in
             if not (check_modulo q challenge && check_modulo q response) then
               raise Exit;
-            let nbeta = cS.beta / (g **~ Z.of_int (min+i-1)) in
+            let g' = if min+i-1 = 0 then G.one else g **~ Z.of_int (min+i-1) in
+            let nbeta = cS.beta / g' in
             let j = 2*i in
             commitments.(j) <- g **~ response *~ cS.alpha **~ challenge;
             commitments.(j+1) <- y **~ response *~ nbeta **~ challenge;
@@ -420,7 +380,8 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
 
   let make_d min max =
     let n = max - min + 1 in
-    let d = Array.make n (invert (g **~ Z.of_int min)) in
+    let g' = if min = 0 then G.one else g **~ Z.of_int min in
+    let d = Array.make n (invert g') in
     for i = 1 to n-1 do
       d.(i) <- d.(i-1) *~ invg
     done;
@@ -484,10 +445,10 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
        let blank_proof = None in
        return {choices; individual_proofs; overall_proof; blank_proof}
 
-  let make_randomness e =
+  let make_randomness () =
     sswap (Array.map (fun q ->
       Array.init (question_length q) (fun _ -> random G.q)
-    ) e.e_params.e_questions)
+    ) election.e_params.e_questions)
 
   let make_sig_prefix zkp commitment =
     "sig|" ^ zkp ^ "|" ^ G.to_string commitment ^ "|"
@@ -503,8 +464,8 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
       ) (Array.to_list answers)
     ) |> Array.of_list
 
-  let create_ballot e ?sk r m =
-    let p = e.e_params in
+  let create_ballot ?sk r m =
+    let p = election.e_params in
     let sk, zkp =
       match sk with
       | None -> None, ""
@@ -525,7 +486,7 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
     ) >>= fun signature ->
     return {
       answers;
-      election_hash = e.e_fingerprint;
+      election_hash = election.e_fingerprint;
       election_uuid = p.e_uuid;
       signature;
     }
@@ -549,10 +510,10 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
        eg_disj_verify y d zkp a.overall_proof sumc
     | _, _ -> false
 
-  let check_ballot e b =
-    let p = e.e_params in
+  let check_ballot b =
+    let p = election.e_params in
     b.election_uuid = p.e_uuid &&
-    b.election_hash = e.e_fingerprint &&
+    b.election_hash = election.e_fingerprint &&
     let ok, zkp = match b.signature with
       | Some {s_public_key = y; s_challenge; s_response} ->
         let zkp = G.to_string y in
@@ -605,11 +566,10 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
 
   type result = elt Serializable_t.result
 
-  let combine_factors num_tallied encrypted_tally partial_decryptions =
-    let dummy = Array.mmap (fun _ -> G.one) encrypted_tally in
-    let factors = Array.fold_left (fun a b ->
-      Array.mmap2 ( *~ ) a b.decryption_factors
-    ) dummy partial_decryptions in
+  type combinator = factor list -> elt array array
+
+  let compute_result num_tallied encrypted_tally partial_decryptions combinator =
+    let factors = combinator partial_decryptions in
     let results = Array.mmap2 (fun {beta; _} f ->
       beta / f
     ) encrypted_tally factors in
@@ -630,21 +590,17 @@ module MakeElection (G : GROUP) (M : RANDOM) = struct
     let result = Array.mmap log results in
     {num_tallied; encrypted_tally; partial_decryptions; result}
 
-  let check_result pks r =
+  let check_result combinator r =
     let {encrypted_tally; partial_decryptions; result; _} = r in
     check_ciphertext encrypted_tally &&
-    (* decryption factors may be not in the same order as pks! *)
-    Array.forall (fun pk ->
-      Array.exists (check_factor encrypted_tally pk) partial_decryptions
-    ) pks &&
-    let dummy = Array.mmap (fun _ -> G.one) encrypted_tally in
-    let factors = Array.fold_left (fun a b ->
-      Array.mmap2 ( *~ ) a b.decryption_factors
-    ) dummy partial_decryptions in
+    let factors = combinator partial_decryptions in
     let results = Array.mmap2 (fun {beta; _} f ->
       beta / f
     ) encrypted_tally factors in
-    Array.fforall2 (fun r1 r2 -> r1 =~ g **~ Z.of_int r2) results result
+    Array.fforall2 (fun r1 r2 ->
+        let g' = if r2 = 0 then G.one else g **~ Z.of_int r2 in
+        r1 =~ g'
+      ) results result
 
   let extract_tally r = r.result
 end
