@@ -1,6 +1,7 @@
 module type PARAMS = sig
   val election : string
   val get_public_keys : unit -> string array option
+  val get_threshold : unit -> string option
   val get_public_creds : unit -> string Stream.t option
   val get_ballots : unit -> string Stream.t option
   val get_result : unit -> string option
@@ -10,7 +11,8 @@ end
 module type S = sig
   val vote : string option -> int array array -> string
   val decrypt : string -> string
-  val finalize : string array -> string
+  val tdecrypt : string -> string -> string
+  val finalize : string list -> string
   val verify : unit -> unit
 end
 
